from anastruct import SystemElements
import numpy as np
L = 1000
F = -2000

struct = SystemElements()
# struct.plotter = CustomPlotter(struct, mesh=50)

struct.add_multiple_elements([[0, 0], [L, 0]], 2)

struct.add_support_hinged(1)
struct.add_support_roll(2)

struct.point_load(3, Fy=F)

struct.q_load(-50, element_id=np.array([1]))

struct.solve()

struct.show_structure()
# struct.show_displacement()
# axes = struct.plotter.plot_structure(figsize=(12, 8),
#         verbosity=0, show=False).get_axes()
# struct.show_shear_force()
struct.show_bending_moment()
# struct.show_reaction_force()

struct.show_results()


# shear_xy = struct.show_shear_force(values_only=True, factor=1)
#
# # Getting real internal forces values for 1 element
# normal_force = struct.get_element_results(element_id=1, verbose=True)['N']
# shear = struct.get_element_results(element_id=1, verbose=True)['Q']
moment = struct.get_element_results(element_id=1, verbose=True)['M']

moment = struct.get_element_result_range(unit = "moment")
# print(moment)

print(np.array([dict['M'] for dict in struct.get_element_results(verbose=True)]).flatten())


'''
Created on May 30, 2018

@author: rch
'''

from scipy import interpolate as ip
from ibvpy.view.plot2d import Viz2D, Vis2D
from scipy.integrate import cumtrapz
import numpy as np
import traits.api as tr

class ForceDisplacement(Vis2D):

    tstep = tr.WeakRef
    Fw = tr.Tuple()

    def _Fw_default(self):
        return ([0], [0])

    def update(self):
        model = self.tstep
        record_dofs = model.control_bc.dofs
        U_ti = model.hist.U_t
        F_ti = model.hist.F_t
        F = np.sum(F_ti[:, record_dofs], axis=1)
        w = U_ti[:, record_dofs[0]]
        self.Fw = F, w


class Vis2DEnergy(Vis2D):

    U_bar_t = tr.List()
    Y_Em_t = tr.List()

    tstep = tr.WeakRef

    def setup(self):
        self.U_bar_t = []
        self.Y_Em_t = []

    def update(self):
        tstep = self.tstep
        mats = tstep.mm
        xdomain = tstep.xdomain
        U = tstep.U_k
        t = tstep.t_n1
        fets = xdomain.fets
        n_c = fets.n_nodal_dofs
        U_Ia = U.reshape(-1, n_c)
        U_Eia = U_Ia[xdomain.I_Ei]
        d = xdomain.integ_factor  # thickness
        eps_Emab = np.einsum(
            'Eimabc,Eic->Emab',
            xdomain.B_Eimabc, U_Eia
        )
        sig_Emab, _ = mats.get_corr_pred(
            eps_Emab, t, **tstep.fe_domain[0].state_n)
        w_m = fets.w_m
        det_J_Em = xdomain.det_J_Em
        U_bar = d / 2.0 * np.einsum(
            'm,Em,Emab,Emab',
            w_m, det_J_Em, sig_Emab, eps_Emab
        )
        self.U_bar_t.append(U_bar)

        Y_Em_t = d / 2.0 * np.einsum(
            'm,Em,Emab,abcd,Emcd->Em',
            w_m, det_J_Em, eps_Emab, mats.D_abcd, eps_Emab
        )
        self.Y_Em_t.append(Y_Em_t)

    def get_t(self):
        return self.tstep.hist.t

    def get_w(self):
        _, w = self.tstep.hist['Fw'].Fw
        return w

    def get_W_t(self):
        P, w = self.tstep.hist['Fw'].Fw
        return cumtrapz(P, w, initial=0)

    G_omega_t = tr.Property
    def _get_G_omega_t(self):
        tstep = self.tstep
        Y_Em_t = np.array(self.Y_Em_t, dtype=np.float_)
        hist = tstep.hist
        t = hist.t
        omega_Em_t = np.array([state[0]['omega'] for state in hist.state_vars])
        # G_omega_Em_t = cumtrapz(Y_Em_t, omega_Em_t, axis=0, initial = 0)
        # return np.sum(G_omega_Em_t, axis=(1,2))

        d_omega_Em_t = (
            (omega_Em_t[1:] - omega_Em_t[:-1]) / (t[1:] - t[:-1])[:,None,None]
        )
        dG_omega_t = np.hstack([
            np.einsum('tEm,tEm->t', Y_Em_t[:-1], d_omega_Em_t), [0]])
        return cumtrapz(dG_omega_t, t, initial=0)

    G_t = tr.Property

    def _get_G_t(self):
        U_bar_t = np.array(self.U_bar_t, dtype=np.float_)
        W_t = self.get_W_t()
        G = W_t - U_bar_t
        return G

    def get_dG_t(self):
        t = self.get_t()
        G = self.get_G_t()
        tck = ip.splrep(t, G, s=0, k=1)
        return ip.splev(t, tck, der=1)


class Vis2DCrackBand(Vis2D):

    tstep = tr.WeakRef

    X_E = tr.Array(np.float_)
    eps_t = tr.List
    sig_t = tr.List
    a_t = tr.List

    def setup(self):
        self.X_E = []
        self.eps_t = []
        self.sig_t = []
        self.a_t = []
        tstep = self.tstep
        xdomain = tstep.xdomain
        crack_band = xdomain.mesh[0, :]  # bt.n_a + 1:]
        E = crack_band.elems
        X = crack_band.dof_X
        X_E = np.average(X[:, :, 1], axis=1)
        self.X_E = X_E

    def update(self):
        tstep = self.tstep
        U = tstep.U_k
        t = tstep.t_n1
        xdomain = tstep.xdomain
        mats = tstep.mm
        fets = xdomain.fets
        n_c = fets.n_nodal_dofs
        U_Ia = U.reshape(-1, n_c)
        U_Eia = U_Ia[xdomain.I_Ei]
        eps_Enab = np.einsum(
            'Einabc,Eic->Enab', xdomain.B_Eimabc, U_Eia
        )
        sig_Enab, _ = mats.get_corr_pred(
            eps_Enab, t, ** tstep.fe_domain[0].state_n
        )
        crack_band = xdomain.mesh[0, :]  # bt.n_a + 1:]
        E = crack_band.elems
        eps_Ey = eps_Enab[E, :, 0, 0]
        eps_E1 = np.average(eps_Ey, axis=1)
        sig_Ey = sig_Enab[E, :, 0, 0]
        sig_E1 = np.average(sig_Ey, axis=1)
        kappa_0 = mats.omega_fn_.kappa_0
        eps_thr = eps_E1 - kappa_0
        a_idx = np.argwhere(eps_thr < 0)
        if len(a_idx) > 0:
            x_1 = self.X_E[0]
            x_a = self.X_E[a_idx[0][0]]
            self.a_t.append(x_a - x_1)
        else:
            self.a_t.append(0)

        self.eps_t.append(eps_E1)
        self.sig_t.append(sig_E1)

    def get_t(self):
        return np.array(self.tstep.hist.t, dtype=np.float_)

    def get_a_x(self):
        return self.X_E

    def get_eps_t(self, t_idx):
        return self.eps_t[t_idx]

    def get_sig_t(self, t_idx):
        return self.sig_t[t_idx]

    def get_a_t(self):
        return np.array(self.a_t, dtype=np.float_)

    def get_da_dt(self):
        a = self.get_a_t()
        t = self.get_t()
        tck = ip.splrep(t, a, s=0, k=1)
        return ip.splev(t, tck, der=1)




import pickle
import json
import os
import pickle

import bmcs_utils.api as bu
import matplotlib.pyplot as plt
import numpy as np
import traits.api as tr
from bmcs_beam.api import BoundaryConfig, DeflectionProfile
from bmcs_cross_section.api import MKappa, EC2, ReinfLayer
from matplotlib.ticker import PercentFormatter
import math
import copy

plt.rcParams["font.family"] = "Times New Roman"
plt.rcParams["font.size"] = 15
# To have math like LaTeX
plt.rcParams['mathtext.fontset'] = 'cm'
plt.rcParams['mathtext.rm'] = 'serif'

class BeamSLSCurve(bu.Model):
    name = 'Beam SLS Curve'

    '''
    - link to data cache identifying the directory where to store the
    interim results
    - dictionary based storage of the individual runs of the study
      which makes it introspective.
    '''

    sls_to_uls_ratio = bu.Float(0.51)
    slenderness_min = bu.Int(3)
    slenderness_max = bu.Int(50)
    rho_min = bu.Float(0.0002)
    rho_max = bu.Float(0.025)
    b = bu.Float(1000)
    h = bu.Float(300)
    high_kappa = bu.Float(0.00007)
    n_kappa = bu.Int(250)

    dp = bu.Instance(DeflectionProfile)

    def _dp_default(self):
        return self._get_dp_and_update_dp_design()

    dp_design = bu.Instance(DeflectionProfile)
    def _dp_design_default(self):
        return self.dp_design

    tree = ['dp', 'dp_design']
    # tree = ['dp']

    n_i = bu.Int(3)
    dense_quarter = bu.Bool(False)

    rho_range = tr.Property(depends_on='rho_max, rho_max, n_i, dense_quarter')

    @tr.cached_property
    def _get_rho_range(self):
        if self.dense_quarter:
            return self._get_range_with_dense_part(self.rho_min, self.rho_max)
        else:
            return np.linspace(self.rho_min, self.rho_max, self.n_i)

    slenderness_range = tr.Property(depends_on='slenderness_min, slenderness_max, n_i, dense_quarter')

    @tr.cached_property
    def _get_slenderness_range(self):
        if self.dense_quarter:
            return self._get_range_with_dense_part(self.slenderness_min, self.slenderness_max)
        else:
            return np.linspace(self.slenderness_min, self.slenderness_max, self.n_i)

    def _get_range_with_dense_part(self, min, max, part=0.25, weight=0.5):
        n_i = self.n_i
        dense_until = min + (max - min) * part
        first_quarter = np.linspace(min, dense_until, int(n_i * weight) + 1)
        rest = np.linspace(dense_until, max, math.ceil(n_i * (1 - weight)))[1:]
        return np.concatenate((first_quarter, rest))

    # pseudo time vars
    interrupt = bu.Bool(False)
    n_seg = bu.Int(5, TIME=True)
    seg = bu.Int(0)

    # dp properties:
    apply_material_factors = bu.Bool(False)
    system_type = bu.Str('dist')  # system_type can be '4pb' or '3pb' or 'dist'
    concrete_law = bu.Str('EC2')  # or 'piecewise linear' or 'EC2 with plateau'
    f_ck = bu.Float(103-8) # Concrete C3, f_cm = 103
    rein_type = bu.Str('steel') # can be 'carbon_grid', 'carbon_rebars', 'carbon_elg'
    use_f_ctm_fl = bu.Bool(False)

    rho = [] # final rho array (corresponds to contour on level 0)
    sl = [] # final slenderness array (corresponds to contour on level 0)
    rho_slider = bu.Float(0.01)
    ld_slider = bu.Float(10)
    indicator_corresponds_to_ld_slider = True

    @tr.observe('rho_slider')
    def _update_dp_according_to_rho(self, event):
        if len(self.rho) != 0:
            self._update_dp_according_to_slider(self.dp)
            self._update_dp_according_to_slider(self.dp_design)
        else:
            print('Generate l/d curves first!')
        self.indicator_corresponds_to_ld_slider = True

    @tr.observe('ld_slider')
    def _update_dp_according_to_ld(self, event):
        d = self.dp.mc.cross_section_shape_.H - list(self.dp.mc.cross_section_layout.items.values())[0].z
        self.dp.beam_design.system_.L = self.ld_slider * d
        self.dp_design.beam_design.system_.L = self.ld_slider * d
        self.indicator_corresponds_to_ld_slider = False

    def _update_dp_according_to_slider(self, dp):
        bd = dp.mc.get_bd()
        A_s = self.rho_slider * bd
        list(dp.mc.cross_section_layout.items.values())[0].A = A_s

        sl = np.interp(self.rho_slider, self.rho, self.sl)
        # TODO, this works only if we have one reinf layer AT THE BOTTOM!
        d = dp.mc.cross_section_shape_.H - list(dp.mc.cross_section_layout.items.values())[0].z
        dp.beam_design.system_.L = sl * d
        dp.mc.state_changed = True

    @tr.observe('apply_material_factors')
    @tr.observe('system_type')
    @tr.observe('concrete_law')
    @tr.observe('f_ck')
    @tr.observe('f_ctm') # Changes on f_ctm (a Property) are not triggering this!
    @tr.observe('rein_type')
    @tr.observe('use_f_ctm_fl')
    def _update_dp(self, event):
        self.dp = self._get_dp_and_update_dp_design()

    ipw_view = bu.View(
        bu.Item('rho_slider',
                editor=bu.FloatRangeEditor(label=r'$\rho$', low=0, high=0.025, n_steps=100, continuous_update=False)),
        bu.Item('ld_slider',
                editor=bu.FloatRangeEditor(label=r'$l/d$', low=1, high=50, n_steps=101, continuous_update=False)),
        bu.Item('slenderness_min', latex=r'{l/d}_\mathrm{min}'),
        bu.Item('slenderness_max', latex=r'{l/d}_\mathrm{max}'),
        bu.Item('rho_min', latex=r'\rho_\mathrm{min}'),
        bu.Item('rho_max', latex=r'\rho_\mathrm{max}'),
        bu.Item('n_i', latex='n_i'),
        bu.Item('dense_quarter', latex=r'\mathrm{Dense~quarter}'),
        bu.Item('apply_material_factors', latex=r'\mathrm{Apply}~\gamma'),
        bu.Item('sls_to_uls_ratio', latex=r'F_\mathrm{SLS}/F_\mathrm{ULS}'),
        bu.Item('system_type', latex=r'\mathrm{System}'),
        bu.Item('concrete_law', latex=r'\mathrm{concrete~law}'),
        bu.Item('f_ck', latex=r'f_{ck}'),
        bu.Item('f_ctm', latex=r'f_{ctm}', editor=bu.FloatEditor()),
        bu.Item('rein_type', latex=r'\mathrm{Rein~type}'),
        # bu.Item('slenderness_range', latex='l/d', editor=bu.IntRangeEditor(value=(10, 35), low_name='slenderness_low', high_name='slenderness_high', n_steps_name='')),
        time_editor=bu.ProgressEditor(
            run_method='run',
            reset_method='reset',
            interrupt_var='interrupt',
            time_var='seg',
            time_max='n_seg'
        )
    )

    _f_ctm = None
    f_ctm = tr.Property(desc='Concrete tensile strength')
    def _set_f_ctm(self, value):
        self._f_ctm = value

        # TODO: this update is a quick fix! tr.observe for f_ctm is not working to update dp
        self.dp = self._get_dp_and_update_dp_design()

    def _get_f_ctm(self):
        if self._f_ctm is not None:
            return self._f_ctm
        else:
            return EC2.get_f_ctm(self.f_ck)

    def _get_dp(self):
        print('dp updated!')
        b = self.b
        h = self.h
        d = 0.9 * h
        f_ck = self.f_ck
        rein_type = self.rein_type

        E = EC2.get_E_cm(f_ck)
        f_ctm_fl = EC2.get_f_ctm_fl(f_ck, h)

        f_ct = f_ctm_fl if self.use_f_ctm_fl else self.f_ctm
        # Unfactor tensile strength (like El-Gha, where f_ctm is used)
        # eps_cr = (1.5/0.85) * f_ct / E if self.apply_material_factors else f_ct / E

        # TODO: check this, it might give bigger f_ctm due to using E_cm
        eps_cr = f_ct / E

        mc = MKappa(low_kappa=0, high_kappa=self.high_kappa, n_kappa=self.n_kappa)
        mc.cs_design.concrete = self.concrete_law
        mc.cs_design.concrete_.trait_set(
            factor=0.85 / 1.5 if self.apply_material_factors else 1,
            eps_cr=eps_cr,
            eps_tu=eps_cr,
            mu=0.0,
        )

        if self.concrete_law == 'EC2 with plateau' or self.concrete_law == 'EC2':
            mc.cs_design.concrete_.trait_set(f_cm=EC2.get_f_cm(f_ck))
        elif self.concrete_law == 'piecewise linear':
            mc.cs_design.concrete_.trait_set(
                E_cc=E,
                E_ct=E,
                eps_cy=EC2.get_eps_c3(f_ck),
                eps_cu=EC2.get_eps_cu3(f_ck),
            )

        # # The default uses f_ctm. Here, I will use f_ctm_fl (in EC2, they tested with both)
        # mc.cs_design.concrete_.eps_cr = (EC2.get_f_ctm_fl(f_ck, h) * 1.5 / 0.85) / mc.cs_design.concrete_.E_ct
        # mc.cs_design.concrete_.eps_tu = (EC2.get_f_ctm_fl(f_ck, h) * 1.5 / 0.85) / mc.cs_design.concrete_.E_ct

        mc.cross_section_shape_.B = b
        mc.cross_section_shape_.H = h

        # T-section
        # mc.cross_section_shape = 'I-shape'
        # mc.cross_section_shape_.H = 200
        # mc.cross_section_shape_.B_w = 50
        # mc.cross_section_shape_.B_f_bot = 50
        # mc.cross_section_shape_.B_f_top = 150
        # mc.cross_section_shape_.H_f_bot = 50
        # mc.cross_section_shape_.H_f_top = 50

        rho = 0.01
        A_s = rho * b * d

        if rein_type == 'steel':
            bl1 = ReinfLayer(name=rein_type, z=h - d, A=A_s, matmod=rein_type)
            bl1.matmod_.trait_set(E_s=200000, f_sy=550, factor=1 / 1.15 if self.apply_material_factors else 1)
        elif rein_type == 'carbon_grid':
            bl1 = ReinfLayer(name=rein_type, z=h - d, A=A_s, matmod='carbon')
            # carbon material factors :
            # alpha_ft * alpha_f_eff / gamma_frp (see El-Ghadioui2020_PhD P. 122) # factor = (char->design) * (mean->char)
            bl1.matmod_.trait_set(E=230000, f_t=3300, factor=(0.85 * 0.9 / 1.3) * (0.85) if self.apply_material_factors else 1)
        elif rein_type == 'carbon_elg':
            bl1 = ReinfLayer(name=rein_type, z=h - d, A=A_s, matmod='carbon')
            # carbon material factors :
            # alpha_ft * alpha_f_eff / gamma_frp (see El-Ghadioui2020_PhD P. 122) # factor = (char->design) * (mean->char)
            f_tm = 1000 * 1.3 / (0.85 * 0.85 * 0.9)
            bl1.matmod_.trait_set(E=100000, f_t= f_tm, factor=(0.85 * 0.9 / 1.3) * (0.85) if self.apply_material_factors else 1)
        elif rein_type == 'carbon_rebars':
            bl1 = ReinfLayer(name=rein_type, z=h - d, A=A_s, matmod='carbon')
            # carbon material factors :
            # alpha_ft * alpha_f_eff / gamma_frp (see El-Ghadioui2020_PhD P. 122)
            bl1.matmod_.trait_set(E=158000, f_t=2500, factor=0.85 * 0.9 / 1.3 if self.apply_material_factors else 1)
        mc.cross_section_layout.add_layer(bl1)
        # mc.state_changed = True

        dp = DeflectionProfile(mc=mc, n_load_steps=100)
        if self.system_type == '4pb':
            dp.beam_design.system = '4pb'
        elif self.system_type == '3pb':
            dp.beam_design.system = '3pb'
        elif self.system_type == 'dist':
            dp.beam_design.system = 'simple_beam_dist_load'

        dp.beam_design.system_.L = 8 * d

        return dp

    def _get_dp_and_update_dp_design(self):
        dp1 = self._get_dp()
        dp2 = self._get_dp()
        self._update_dp_design(dp2)
        return dp1

    def _update_dp_design(self, dp_design):
        f_ck = self.f_ck

        E_k = f_ck / EC2.get_eps_c3(f_ck)

        if f_ck < 100:
            dp_design.mc.cs_design.concrete = 'EC2 with plateau'
            dp_design.mc.cs_design.concrete_.factor = 0.85 / 1.5
            dp_design.mc.cs_design.concrete_.f_cm = f_ck + 8
        else:
            dp_design.mc.cs_design.concrete = 'piecewise linear'
            dp_design.mc.cs_design.concrete_.trait_set(
                factor=0.85 / 1.5,
                E_cc=E_k,
                E_ct=E_k,
                eps_cy=EC2.get_eps_c3(f_ck),
                eps_cu=EC2.get_eps_cu3(f_ck),
            )

        # f_ctm_fl = EC2.get_f_ctm_fl(f_ck, dp_design.mc.cross_section_shape_.H)
        # f_ctm = f_ctm_fl if self.use_f_ctm_fl else self.f_ctm
        # f_ctk = EC2.get_f_ctk_0_05(f_ck)
        # eps_cr = f_ctk / E
        # dp_design.mc.cs_design.concrete_.trait_set(
        #     eps_cr=eps_cr,
        #     eps_tu=eps_cr,
        #     mu=0.0,
        # )

        for reinf_layer in list(dp_design.mc.cross_section_layout.items.values()):
            reinf_type = reinf_layer.name
            if reinf_type == 'steel':
                reinf_layer.matmod_.factor = (1 / 1.15) * (1 / 1.1) # (char->design) * (mean->char)
            else:
                # carbon material factors :
                # alpha_ft * alpha_f_eff / gamma_frp (see El-Ghadioui2020_PhD P. 122)
                reinf_layer.matmod_.factor = (0.85 * 0.9 / 1.3) * (0.85) # (char->design) * (mean->char)

        self.dp_design = dp_design

    def run(self, update_progress=lambda t: t):
        print('run started...')
        F_u_grid, F_s_grid, rho_grid, sl_grid = self.get_Fu_and_Fs()
        self.plot_with_ec2_curves(F_u_grid, F_s_grid, rho_grid, sl_grid,
                                  self.ax1 if hasattr(self, 'ax1') else None)
        print('run finished...')

    def reset(self):
        # reset everything to default values
        pass

    def subplots(self, fig):
        self.ax1, self.ax2 = fig.subplots(2, 1)
        return self.ax1, self.ax2

    def update_plot(self, axes):
        sl_curve_ax, ld_ax = axes
        if len(self.rho) != 0:
            if self.indicator_corresponds_to_ld_slider:
                sl = np.interp(self.rho_slider, self.rho, self.sl)
                sl_curve_ax.plot(self.rho_slider, sl, color='orange', marker='o')
            else:
                sl_curve_ax.plot(self.rho_slider, self.ld_slider, color='orange', marker='o')
        if hasattr(self, 'F_u_grid'):
            if len(self.F_u_grid) != 0:
                self.plot_with_ec2_curves(self.F_u_grid, self.F_s_grid, self.rho_grid, self.sl_grid, sl_curve_ax)
        self.plot_ld_curves(ld_ax)

    def plot_ld_curves(self, ax):
        eta = self.sls_to_uls_ratio
        self.dp.w_SLS = True

        self.dp.plot_fw_with_fmax(ax, f_max_label=r'$F_\mathrm{mean, u}$', f_max_color='black')
        self.dp_design.plot_fw_with_fmax(ax, f_max_label=r'$F_\mathrm{design, u}$')
        F_ULS_design = abs(self.dp_design.final_plot_F_scale * self.dp_design.beam_design.system_.F)
        F_SLS_design = eta * F_ULS_design

        ax.axhline(y=F_SLS_design, linestyle='--', color='r')
        ax.annotate(r'$F_{\mathrm{design,~SLS,~qp}} = $' + str(round(F_SLS_design, 2)), xy=(0, 1.06 * F_SLS_design),
                    color='r')

    # not used
    def save(self):
        out_dir = os.path.join(bu.data_cache.dir, self.__class__.__name__)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        with open(os.path.join(out_dir, 'class'), 'wb') as out_file:
            """ Giving TypeError: cannot pickle 'weakref' object"""
            pickle.dump(self, out_file)

    # not used
    def save_data_vars_in_json(self, f_ck, dp, path):
        mc = dp.mc
        rein = list(mc.cross_section_layout.items.values())
        output_data = {'mc.n_m': mc.n_m,
                       'mc.n_kappa': mc.n_kappa,
                       'mc.low_kappa': mc.low_kappa,
                       'mc.high_kappa': mc.high_kappa,
                       'mc.E_cc': mc.cs_design.concrete_.E_cc,
                       'mc.E_ct': mc.cs_design.concrete_.E_ct,
                       'mc.eps_tu': mc.cs_design.concrete_.eps_tu,
                       'mc.eps_cr': mc.cs_design.concrete_.eps_cr,
                       'mc.eps_cy': mc.cs_design.concrete_.eps_cy,
                       'mc.eps_cu': mc.cs_design.concrete_.eps_cu,
                       'mc.mu': mc.cs_design.concrete_.mu,
                       'f_ck': f_ck,
                       'rein[0].E': rein[0].matmod_.E,
                       'rein[0].z': rein[0].z,
                       'rein[0].A': rein[0].A,
                       'dp.beam_design.system_.L': dp.beam_design.system_.L, }
        with open(path, 'w') as outfile:
            json.dump(output_data, outfile, sort_keys=True, indent=4)

    def get_Fu_and_Fs(self, upper_reinforcement=False, plot=False):
        self.F_u_design_grid, self.F_s_design_grid, _, _, _, _ = self._get_grids(self.dp_design, calc_shear=False, mean=False)
        self.F_SLS_qp_design_grid = self.sls_to_uls_ratio * self.F_u_design_grid
        dp_grids = self._get_grids(self.dp, calc_shear=True, mean=True)
        self.F_u_grid, self.F_s_grid, self.rho_grid, self.sl_grid, self.F_u_shear_grid, self.w_SLS_qp_grid = dp_grids

        self.F_u_design_to_mean_grid = self.F_u_design_grid / self.F_u_grid
        return self.F_u_grid, self.F_s_grid, self.rho_grid, self.sl_grid


    def _get_grids(self, dp, upper_reinforcement=False, plot=False, calc_shear=False, mean=True):

        slenderness_range = self.slenderness_range
        rho_range = self.rho_range

        if upper_reinforcement:
            d = list(dp.mc.cross_section_layout.items.values())[0].z
        else:
            d = dp.mc.cross_section_shape_.H - list(dp.mc.cross_section_layout.items.values())[0].z

        bd = dp.mc.get_bd(upper_reinforcement = upper_reinforcement)

        rho_grid, sl_grid = np.meshgrid(rho_range, slenderness_range)
        F_u_grid = np.zeros_like(rho_grid)
        F_s_grid = np.zeros_like(rho_grid)
        w_SLS_qp_grid = np.zeros_like(rho_grid)
        F_u_shear_grid = np.zeros_like(rho_grid)

        if plot:
            _, ax = plt.subplots()
            ax.set_xlabel(r'$w$ [mm]')
            ax.set_ylabel(r'$F$ [KN]')

        for sl_idx in range(len(slenderness_range)):
            for rho_idx in range(len(rho_range)):
                if self.interrupt:
                    return

                rho = rho_grid[rho_idx, sl_idx]
                sl = sl_grid[rho_idx, sl_idx]

                print('parameter combination', rho, sl)

                # assigning the grid area (area_g) to the reinforcement area variable
                A_j_g = rho * bd
                list(dp.mc.cross_section_layout.items.values())[0].A = A_j_g

                # assigning the grid length (L_g) to the beam length variable
                L_g = sl * d
                dp.beam_design.system_.L = L_g

                dp.mc.state_changed = True

                # running the deflection analysis
                F_data, w_data = dp.get_Fw()
                F_data = F_data * dp.final_plot_F_scale

                # plotting, post-processing & saving the data
                if plot:
                    ax.plot(w_data, F_data, label="rho={}%-sl={} ".format(rho * 100, sl))

                w_s = dp.beam_design.system_.L / 250
                F_u = max(F_data)
                F_s = np.interp(w_s, w_data, F_data, right=F_u * 2)

                if calc_shear and dp.shear_force_can_be_calculated():
                    F_u_shear_grid[rho_idx, sl_idx] = dp.get_nm_shear_force_capacity()
                F_u_grid[rho_idx, sl_idx] = F_u
                F_s_grid[rho_idx, sl_idx] = F_s

                if mean:
                    F_SLS = self.sls_to_uls_ratio * self.F_u_design_grid[rho_idx, sl_idx]
                    w_SLS_qp_grid[rho_idx, sl_idx] = np.interp(F_SLS, F_data, w_data, right=max(w_data) * 2)

        return F_u_grid, F_s_grid, rho_grid, sl_grid, F_u_shear_grid, w_SLS_qp_grid

    def plot_with_ec2_curves(self, F_u_grid, F_s_grid, rho_grid, sl_grid, ax=None, label=None):
        if not ax:
            fig, ax = plt.subplots()

        color = np.random.rand(3, )

        z = F_u_grid / F_s_grid - 1. / (self.sls_to_uls_ratio * self.F_u_design_to_mean_grid)
        cs = ax.contour(rho_grid, sl_grid, z, levels=[0], colors=[color])
        if label:
            cs.collections[0].set_label(label)

        # Put values on lines
        # ax.clabel(cs, inline=True, fontsize=10)

        p = cs.collections[0].get_paths()[0]
        v = p.vertices
        self.rho = v[:, 0]
        self.sl = v[:, 1]

        # Draw EC2 curve
        self.plot_steel_sls_curves(ax, f_cks=[self.f_ck],
                                   axes_start_from_zero=True,
                                   color=color)

    def plot_steel_sls_curves(self, ax=None,
                              rho_range=np.linspace(0.0025, 0.025, 1000),
                              f_cks=np.arange(20, 110, 10),
                              rho_p=0,
                              K=1,
                              axes_start_from_zero=False,
                              color=None):
        if not ax:
            fig, ax = plt.subplots()

        slenderness = []
        for f_ck in f_cks:
            for rho in rho_range:
                slenderness.append(BeamSLSCurve.get_slenderness_limit(rho, f_ck, rho_p, K))
            ax.plot(rho_range, slenderness, label=r'$f_{ck} = $' + str(f_ck) + ' MPa (Steel - EC2)', ls='--', c=color)

            slenderness = []

        if axes_start_from_zero:
            ax.set_ylim(0, self.slenderness_max)
            ax.set_xlim(0, rho_range[-1])
        else:
            ax.set_ylim(10, self.slenderness_max)
            ax.set_xlim(rho_range[0], rho_range[-1])

        ax.xaxis.set_major_formatter(PercentFormatter(xmax=1))
        ax.set_ylabel(r'$l/(K \cdot d$)')
        ax.set_xlabel(r'Tensile reinforcement ratio $\rho$ [%]')
        ax.set_title(r'$l/d$ limits according to EC2 eqs. 7.16a & 7.16b')
        ax.grid(color='#e6e6e6', linewidth=0.7)
        ax.legend()

        if not ax:
            fig.show()

    @staticmethod
    def get_slenderness_limit(rho, f_ck, rho_p=0, K=1):
        # see EC2, see eqs 7.16
        rho_0 = 0.001 * np.sqrt(f_ck)
        if rho <= rho_0:
            return K * (11 + 1.5 * ((f_ck) ** 0.5) * (rho_0 / rho) + 3.2 * ((f_ck) ** 0.5) * (
                    (rho_0 / rho - 1) ** (3 / 2)))
        else:
            return K * (11 + 1.5 * ((f_ck) ** 0.5) * (rho_0 / (rho - rho_p)) + (1 / 12) * (f_ck ** 0.5) * (
                    (rho_p / rho_0) ** 0.5))

    def plot_F_u(self, load='bending', bending_shear_diff=False, scale=1, levels=None, sls=None):
        fig, ax = plt.subplots()
        fig.set_size_inches(5.5, 4.5)

        if sls is not None:
            z = self.F_u_grid/sls.F_u_grid
            levels = np.linspace(0, 5, 11)
            # z = self.F_u_grid - sls.F_u_grid
        elif bending_shear_diff:
            z = self.F_u_shear_grid - self.F_u_grid
        else:
            z = self.F_u_grid * scale if load == 'bending' else self.F_u_shear_grid * scale
        if levels is None:
            levels = [0, 10, 20, 30, 40, 60, 80, 100, 150, 300, 1000, 2000, 3000, 4000, 5000, 6000, 7000, 8000, -10,
                      -20, -30, -40, -60, -80, -100, -150, -300, -1000, -2000, -3000, -4000, -5000, -6000, -7000,
                      -8000]
            levels = np.array(levels) * scale
        levels.sort()
        cs = ax.contour(self.rho_grid, self.sl_grid, z, levels=levels) #cmap='RdYlBu_r', cmap='turbo', cmap='rainbow'
        cs_shade = ax.contourf(self.rho_grid, self.sl_grid, z, levels=[-1e6, 0], colors=['lightgray', 'white'])
        ax.clabel(cs, inline=True, fontsize=14)
        # sls.sls_to_uls_ratio = 0.59
        # sls._plot_with_ec2_curves(0.3*sls.F_u_grid, sls.F_s_grid, sls.rho_grid, sls.sl_grid, ax=ax)

        ax.xaxis.set_major_formatter(PercentFormatter(xmax=1))
        ax.set_ylabel(r'$L/d$')
        ax.set_xlabel(r'Reinforcement ratio $\rho$')
        ax.set_title(r'Contour lines of maximum load $F_\mathrm{max}$ [kN]') # where F_max = ql
        ax.grid(color='#e6e6e6', linewidth=0.7)
        ax.set_ylim(ymin=0)
        ax.set_xlim(xmin=0)
        # ax.legend()

        # fig.show()

        return fig


class SLSParamStudy(bu.ParametricStudy):
    def __init__(self, b_sls):
        self.b_sls = b_sls

    def plot(self, ax, plot_title, param_name_value):
        ax.set_xlabel(r'$w_\mathrm{max}$ [mm]')
        ax.set_ylabel(r'$F$ [kN]')

        self.b_sls.dp.mc.state_changed = True

        F_u_grid, F_s_grid, rho_grid, sl_grid = self.b_sls.get_Fu_and_Fs()
        self.b_sls.plot_with_ec2_curves(F_u_grid, F_s_grid, rho_grid, sl_grid, ax, label=param_name_value)
        ax.set_title(plot_title + ' effect')
        ax.legend()

        out_dir = self.get_output_dir()
        np.save(os.path.join(out_dir, 'F_u_grid' + '_' + self.b_sls.rein_type + '__' + param_name_value + '.npy'),
                F_u_grid)
        np.save(os.path.join(out_dir, 'F_s_grid' + '_' + self.b_sls.rein_type + '__' + param_name_value + '.npy'),
                F_s_grid)
        np.save(os.path.join(out_dir, 'rho_grid' + '_' + self.b_sls.rein_type + '__' + param_name_value + '.npy'),
                rho_grid)
        np.save(os.path.join(out_dir, 'sl_grid' + '_' + self.b_sls.rein_type + '__' + param_name_value + '.npy'),
                sl_grid)

    # # to be updated
    # def plot_all_curves(self):
    #     f_cks = [50]
    #     F_u_grids = []
    #     F_s_grids = []
    #     rho_grids = []
    #     sl_grids = []
    #     reinforcement = 'carbon'
    #     for f_ck in f_cks:
    #         f_ck = str(f_ck)
    #         F_u_grids.append(
    #             np.load('exported_data/F_u_grid_carbon_EC2_eq2_tension_E230_ft_3000_c' + str(f_ck) + '.npy'))
    #         F_s_grids.append(
    #             np.load('exported_data/F_s_grid_carbon_EC2_eq2_tension_E230_ft_3000_c' + str(f_ck) + '.npy'))
    #         rho_grids.append(
    #             np.load('exported_data/rho_grid_carbon_EC2_eq2_tension_E230_ft_3000_c' + str(f_ck) + '.npy'))
    #         sl_grids.append(np.load('exported_data/sl_grid_carbon_EC2_eq2_tension_E230_ft_3000_c' + str(f_ck) + '.npy'))
    #
    #     _, ax = plt.subplots(1, 1)
    #
    #     ax.set_ylabel('L/d')
    #     ax.set_xlabel(r'$\rho$ %')
    #     ax.set_ylim(0, self.slenderness_max)
    #     ax.set_xlim(0.0, self.rho_max)
    #
    #     for f_ck, F_u_grid, F_s_grid, rho_grid, sl_grid in zip(f_cks, F_u_grids, F_s_grids, rho_grids, sl_grids):
    #         z = 0.5 * F_u_grid / F_s_grid - 1. / self.b_sls.sls_to_uls_ratio
    #         CS = ax.contour(rho_grid, sl_grid, z, colors=[np.random.rand(3, )], levels=[0])
    #         CS.collections[0].set_label('C' + str(f_ck))
    #         #     ax.clabel(CS, inline=1, fontsize=10)
    #
    #         BeamSLSCurve().plot_steel_sls_curves(ax=ax, f_cks=[50], axes_start_from_zero=True)
    #
    #     ax.legend()



# elif config == BoundaryConfig.THREE_SPAN_DIST_LOAD:
#     # 3 span distributed load example
#     beam.apply_load(R1, 0, -1)
#     beam.apply_load(R2, L / 3, -1)
#     beam.apply_load(R3, 2 * L / 3, -1)
#     beam.apply_load(R4, L, -1)
#     beam.apply_load(F, 0, 0)
#     beam.bc_deflection = [(0, 0), (L / 3, 0), (2 * L / 3, 0), (L, 0)]
#
# elif config == BoundaryConfig.THREE_PB_FIXED_SUPPORT:
#     # fixed support example
#     beam.apply_load(R1, 0, -1)
#     beam.apply_load(M1, 0, -2)
#     beam.apply_load(R2, L, -1)
#     beam.apply_load(M2, L, -2)
#     beam.apply_load(F, L / 2, -1)
#     beam.bc_deflection = [(0, 0), (L, 0)]
#     beam.bc_slope = [(0, 0), (L, 0)]
#
# elif config == BoundaryConfig.SINGLE_MOMENT:
#     # single moment example
#     beam.apply_load(R1, 0, -1)
#     beam.apply_load(R2, L, -1)
#     beam.apply_load(-F, L / 2, -2)
#     beam.bc_deflection = [(0, 0), (L, 0)]
#
# elif config == BoundaryConfig.FIXED_SUPPORT_DIST_LOAD:
#     beam.apply_load(R1, 0, -1)
#     beam.apply_load(M1, 0, -2)
#     beam.apply_load(R2, L, -1)
#     beam.apply_load(M2, L, -2)
#     beam.apply_load(F, 0, 0)
#     beam.bc_deflection = [(0, 0), (L, 0)]
#     beam.bc_slope = [(0, 0), (L, 0)]
#
# elif config == BoundaryConfig.FIXED_AND_ROLLER_SUPPORT_DIST_LOAD:
#     beam.apply_load(R1, 0, -1)
#     beam.apply_load(R2, L, -1)
#     beam.apply_load(M2, L, -2)
#     beam.apply_load(F, 0, 0)
#     beam.bc_deflection = [(0, 0), (L, 0)]
#     beam.bc_slope = [(L, 0)]


from bmcs_utils.api import Model, Int, Item, View, Float
from bmcs_beam.beam_config.system.beam_system import BeamSystem
import numpy as np

class CantileverDistLoadSystem(BeamSystem):

    name = 'CantileverDistLoadSystem'

    F = Float(-5, SYSTEM=True)
    n_x = 2

    tree = []

    ipw_view = View(
        *BeamSystem.ipw_view.content,
        Item('F', latex='F \mathrm{[kN/m]}', readonly=True), # kN/m = N/mm
    )

    def update_struct(self):
        self.struct = self.get_new_struct()
        self.struct.add_multiple_elements([[0, 0], [self.L, 0]], self.n_x)
        self.struct.add_support_fixed(1)
        elements = np.arange(1, self.n_x + 1)
        self.struct.q_load(q=self.F, element_id=elements)
        self.struct.solve()

    def get_max_force(self, M_I):
        # This should be done numerically, but here for efficiency and
        # because this is a known case it's given directly
        return 2 * np.min(M_I) / self.L ** 2

    def get_plot_force_scale_and_unit(self):
        """ Scale which should be applied on the force when plotting """
        # Total load
        return 1 * self.L/1000, 'kN'
        # kN/m load
        # return 1, 'kN/m'


from bmcs_utils.api import Model, Int, Item, View, IntEditor
from bmcs_beam.beam_config.system.anastruct.anastruct_custom_plotter import CustomPlotter
import traits.api as tr
from anastruct import SystemElements

class System(Model):

    name = 'System'

    n_per_element = Int(50, desc='Postprocessing mesh size per element. Has no influence on the fe calculation.',
                        SYSTEM=True)
    tree = []

    ipw_view = View(
        Item('n_per_element', latex='n_\mathrm{per~elem}', editor=IntEditor(min=10)),
    )

    @tr.observe("+SYSTEM")
    def _update_struct_state(self, event):
        self.update_struct()

    def update_struct(self):
        return NotImplementedError

    def get_new_struct(self):
        struct = SystemElements()
        struct.plotter = CustomPlotter(struct, mesh=self.n_per_element)
        return struct


import numpy as np
from bmcs_utils.api import Model, Int, Item, View, Float, FloatEditor, IntEditor
from bmcs_beam.beam_config.system.beam_system import BeamSystem
import traits.api as tr

class FourPBSystem(BeamSystem):

    name = 'FourPBSystem'

    F = Float(-1000, SYSTEM=True)
    L_F = Float(2000, desc='length from support to the force', SYSTEM=True)
    """ L_F defaults to L/3, if you want to change it, change it AFTER setting L"""

    _n_x_min = 3
    n_x = 3

    tree = []

    @tr.observe('L')
    def update_L_F(self, event):
        self.L_F = self.L/3

    ipw_view = View(
        *BeamSystem.ipw_view.content,
        Item('F', latex='F \mathrm{[N]}', readonly=True),
        Item('L_F', latex='L_F \mathrm{[N]}', editor=FloatEditor(min=4)),
    )

    def update_struct(self):
        self.struct = self.get_new_struct()

        n_x_to_F = max(int((self.L_F / self.L) * self.n_x), 1)
        n_x_between_forces = max(self.n_x - 2 * n_x_to_F, 1)
        if n_x_to_F == 1:
            self.struct.add_element([[0, 0], [self.L_F, 0]])
        else:
            self.struct.add_multiple_elements([[0, 0], [self.L_F, 0]], n_x_to_F)
        if n_x_between_forces == 1:
            self.struct.add_element([[self.L_F, 0], [self.L - self.L_F, 0]])
        else:
            self.struct.add_multiple_elements(location=[[self.L_F, 0], [self.L - self.L_F, 0]], n=n_x_between_forces)
        if n_x_to_F == 1:
            self.struct.add_element([[self.L - self.L_F, 0], [self.L, 0]])
        else:
            self.struct.add_multiple_elements([[self.L - self.L_F, 0], [self.L, 0]], n_x_to_F)

        self.struct.add_support_hinged(1)
        self.struct.add_support_roll(self.n_x + 1)

        self.struct.point_load(n_x_to_F + 1, Fy=self.F)
        self.struct.point_load(n_x_to_F + n_x_between_forces + 1, Fy=self.F)
        self.struct.solve()

    def get_max_force(self, M_I):
        # This should be done numerically, but here for efficiency and
        # because this is a known case it's given directly
        return np.max(M_I) / self.L_F

    def get_plot_force_scale_and_unit(self):
        """ Scale which should be applied on the force when plotting """
        return 2/1000, 'kN'


import traits.api as tr
import numpy as np

from bmcs_utils.api import InteractiveModel, View, Item, Button, ButtonEditor, Float, Int, \
    mpl_align_yaxis_to_zero, mpl_show_one_legend_for_twin_axes, ParametricStudy
from bmcs_utils.api import Model, View, Item, Button, Bool, Float, Int, \
    mpl_align_yaxis, ParametricStudy

import bmcs_utils.api as bu

from bmcs_beam.beam_config.beam_design import BeamDesign
from bmcs_cross_section.mkappa import MKappa
from scipy.integrate import cumulative_trapezoid as cumtrapz
import matplotlib.gridspec as gridspec

from bmcs_beam.beam_config.boundary_conditions import BoundaryConditions, BoundaryConfig
from bmcs_beam.beam_config.system.cantilever_system import CantileverDistLoadSystem
from bmcs_beam.beam_config.system.simple_dist_load_system import SimpleDistLoadSystem


class DeflectionProfile(Model):
    '''
    Deflection model of a BMCS beam
    '''

    name = 'Deflection Profile'

    beam_design = bu.Instance(BeamDesign, ())
    mc = bu.Instance(MKappa, ())
    n_load_steps = Int(31)
    w_SLS = Bool(False)

    depends_on = ['beam_design', 'mc']
    tree = ['beam_design', 'mc']

    ipw_view = View(
        Item('n_load_steps', latex=r'n_{\mathrm{load~steps}}'),
        Item('w_SLS', latex=r'\mathrm{SLS_{limit}}')
    )

    f_exp_data = tr.List
    w_exp_data = tr.List

    plot_F_scale = Float(1, desc='Additional plotting scale that can be added by the user')

    final_plot_F_scale = tr.Property()
    def _get_final_plot_F_scale(self):
        F_scale_to_kN, _ = self.beam_design.system_.get_plot_force_scale_and_unit()
        return self.plot_F_scale * F_scale_to_kN

    F_unit = tr.Property()
    def _get_F_unit(self):
        _, unit = self.beam_design.system_.get_plot_force_scale_and_unit()
        return unit

    def add_fw_exp(self, load_array, deflection_array):
        self.f_exp_data.append(load_array)
        self.w_exp_data.append(deflection_array)

    def get_kappa_x(self):
        '''
        Profile of curvature along the beam
        '''
        M = self.beam_design.get_M_x()
        return self.mc.get_kappa_M(M)
    
    def get_kappa_shrinkage(self):
        '''
        Calculate the shrinkage curvature based on EC2
        '''
        f_ck = 30
        t = 365
        t_s = 3

        alpha_ds1 = 4
        alpha_ds2 = 0.12

        RH = 0.70
        RH_0 = 1.00

        f_cm = 30
        f_cmo = 10
        phi = 2.5
        E_s = 200000
        E_cm = 33000
        S = 1000 * 300 ** 2 / 2
        I = 1000 * 300 ** 3 / 12
        A_c = 1000 * 300
        u = 2 * (1000 + 300)

        eps_ca_infty = 2.5 * (f_ck - 10) * 1e-6
        beta_as_t = 1 - np.exp(- 0.2 * t ** 0.5)
        eps_ca = beta_as_t * eps_ca_infty

        beta_RH = 1.55 * (1 - (RH/RH_0)**3)
        h_0 = 2 * A_c / u
        beta_ds_t_t_s = (t - t_s) / ((t - t_s) + 0.04 * h_0 ** (3/2))
        h_0_ = [100, 200 ,300, 500, 800]
        k_h_ = [ 1, 0.85, 0.75, 0.7, 0.7 ]
        k_h = np.interp(h_0, h_0_, k_h_)
        eps_cd0 = 0.85 * ((220 + 110 * alpha_ds1) * np.exp(-alpha_ds2 * f_cm / f_cmo)) * 1e-6 * beta_RH
        eps_cd = beta_ds_t_t_s * k_h * eps_cd0

        eps_cs = eps_cd + eps_ca

        E_c_eff = E_cm / (1 + phi)
        alpha_e = E_s / E_c_eff
        kappa_cs = eps_cs * alpha_e * S / I

        kappa_cs_ = np.array([kappa_cs])
        kappa_cs_x = np.zeros_like(self.get_kappa_x())
        kappa_cs_x[:] = kappa_cs_

        return kappa_cs_x

    def get_phi_x(self):
        '''
        Calculate the cross sectional rotation by integrating the curvature
        '''
        # TODO rename phi to theta
        kappa_x = self.get_kappa_x() # + 2e-6 #+ self.get_kappa_shrinkage()
        # Kappa = 1/R = d_phi/d_x
        phi_int_x = cumtrapz(kappa_x, self.beam_design.system_.x, initial=0)

        # ----- Applying rotation bcs to resolve integration constant -----
        # ----- rotation = integrated_rotation + c -----
        if not isinstance(self.beam_design.system_, CantileverDistLoadSystem):
                # and not isinstance(self.beam_design.system_, FixedAndRollerDistLoadBeamSystem):
            # The bc for symmetric beams (3PB, 4PB) is (rotation(L/2) = 0 <-> rotation(x_cond_1) = rot_cond_1)
            # -> integrated_rotation(L/2) + c = 0 - > c = -integrated_rotation(L/2) + 0
            # TODO this bc works only for symmetric beams - generalize for other systems
            L = self.beam_design.system_.L
            x_cond_1, rot_cond_1 = (L / 2, 0)
            int_constant = -np.interp(x_cond_1, self.beam_design.system_.x, phi_int_x) + rot_cond_1
            phi_x = phi_int_x + int_constant
        # elif isinstance(self.beam_design.system_, FixedAndRollerDistLoadBeamSystem):
        #     phi_x -= phi_x[-1]

        # This is a more general approach but not sure from it!
        # kappa_derivative = np.gradient(kappa_x, self.beam_design.system_.x)
        # kappa_derivative_signs = np.sign(kappa_derivative)
        # sign_change = (np.diff(kappa_derivative_signs) != 0) * 1
        # if sign_change.size != 1:
        #     print("Warning: multiple curvature slope changes has been found but only the first is considered!!")
        # sign_change_index = np.where(sign_change == 1)[0][0]
        # phi_x -= phi_x[sign_change_index]

        return phi_x

    def get_w_x(self):
        """
        Profile of deflection along the beam
        """
        phi_x = self.get_phi_x()
        w_int_x = cumtrapz(phi_x, self.beam_design.system_.x, initial=0)

        # # ----- Applying deflection bcs to resolve integration constant -----
        # # ----- deflection = integrated_deflection + c (w_x = w_int_x + int_constant) -----
        # # The bc for simply supported beam or simple cantilever we require zero deflection at the
        # # left support - the right one comes automatically
        # # -> integrated_deflection(0) + c = 0 - > c = -integrated_rotation(0)
        # # TODO: generalize for other systems
        # L = self.beam_design.system_.L
        # x_cond_1, w_cond_1 = (0, 0)
        # int_constant = -np.interp(x_cond_1, self.beam_design.system_.x, w_int_x) + w_cond_1
        # w_x = w_int_x + int_constant

        # For efficiency, the following is used as int_constant = 0 in the case above:
        w_x = w_int_x

        # if self.beam_design.beam_conf_name != BoundaryConfig.FIXED_AND_ROLLER_SUPPORT_DIST_LOAD \
        #         and self.beam_design.beam_conf_name != BoundaryConfig.CANTILEVER_DIST_LOAD:
        #     w_x += w_x[0]
        return w_x

    theta_max = Float(1)

    F_max = tr.Property(Float)
    ''''
    Identify the ultimate limit state based on the maximum moment capacity
    of the cross section.
    '''
    def _get_F_max(self):
        M_I, kappa_I = self.mc.inv_M_kappa
        F_max = self.beam_design.system_.get_max_force(M_I)
        return abs(F_max)

    # def run(self):
    #     F_arr = np.linspace(0, self.F_max, self.n_load_steps)
    #     w_list = []
    #     original_F = self.beam_design.system_.F
    #     for F in F_arr:
    #         if F == 0:
    #             w_list.append(0)
    #         else:
    #             self.beam_design.system_.F = -F
    #             # Append the maximum deflection value that corresponds to the new load (F)
    #             w_list.append(np.fabs(np.min(self.get_w_x())))
    #     self.beam_design.system_.F = original_F
    #     return F_arr, np.array(w_list)
    #
    # def reset(self):
    #     self.theta_F = 0

    def _get_F_arr(self, F_max):
        n_load_steps = self.n_load_steps
        # Make F_arr list denser up to (0.2 * the range) to capture cracking load properly, otherwise
        # deflections in SLS might be inaccurate
        n_1 = int(0.4 * n_load_steps)
        n_2 = n_load_steps - n_1
        return np.concatenate((np.linspace(0, 0.2 * F_max, n_1, endpoint=False),
                                   np.linspace(0.2 * F_max, F_max, n_2)))

    F_max_old = Float

    def get_Fw(self):
        F_max = self.F_max
        F_arr = self._get_F_arr(F_max)
        w_list = []
        # @todo [SR,RC]: separate the slider theta_F from the calculation
        #                of the datapoints load deflection curve.
        #                use broadcasting in the functions
        #                get_M_x(x[:,np.newaxis], F[np.newaxis,:] and
        #                in get_Q_x, get_kappa_x, get_w_x, get_phi_x, get_w_x
        #                then, the browsing through the history is done within
        #                the two dimensional array of and now loop over theta is
        #                neeeded then. Theta works just as a slider - as originally
        #                introduced.
        original_F = self.beam_design.system_.F
        for F in F_arr:
            if F == 0:
                w_list.append(0)
            else:
                self.beam_design.system_.F = -F
                # Append the maximum deflection value that corresponds to the new load (F)
                w_list.append(np.max(np.fabs(self.get_w_x())))
        if self.F_max_old == F_max:
            self.beam_design.system_.F = original_F
        self.F_max_old = F_max
        return F_arr, np.array(w_list)
    
    # def get_Fw_inx(self, inx):
    #     F_max = self.F_max
    #     F_arr = np.linspace(0, F_max, self.n_load_steps)
    #     w_list = []
    #     # @todo [SR,RC]: separate the slider theta_F from the calculation
    #     #                of the datapoints load deflection curve.
    #     #                use broadcasting in the functions
    #     #                get_M_x(x[:,np.newaxis], F[np.newaxis,:] and
    #     #                in get_Q_x, get_kappa_x, get_w_x, get_phi_x, get_w_x
    #     #                then, the browsing through the history is done within
    #     #                the two dimensional array of and now loop over theta is
    #     #                neeeded then. Theta works just as a slider - as originally
    #     #                introduced.
    #     original_F = self.beam_design.system_.F
    #     for F in F_arr:
    #         if F == 0:
    #             w_list.append(0)
    #         else:
    #             self.beam_design.system_.F = -F
    #             # Append the maximum deflection value that corresponds to the new load (F)
    #             w_list.append(np.fabs(self.get_w_x()[inx]))
    #     if self.F_max_old == F_max:
    #         self.beam_design.system_.F = original_F
    #     self.F_max_old = F_max
    #     return F_arr, np.array(w_list)


    def subplots(self, fig):
        gs = gridspec.GridSpec(1, 2, figure=fig, width_ratios=[0.7, 0.3])

        # ax2, ax3 = fig.subplots(1, 2)
        ax_w = fig.add_subplot(gs[0, 0])
        ax_k = ax_w.twinx()
        ax_Fw = fig.add_subplot(gs[0, 1])
        return ax_w, ax_k, ax_Fw

    def update_plot(self, axes):
        ax_w, ax_k, ax_Fw = axes
        self.plot_fw_with_fmax(ax_Fw)
        self.plot_curvature_along_beam(ax_k)
        self.plot_displacement_along_beam(ax_w)
        mpl_align_yaxis_to_zero(ax_w, ax_k)
        mpl_show_one_legend_for_twin_axes(ax_w, ax_k)

    def plot_fw_with_fmax(self, ax_Fw, f_max_label = r'$F_{\mathrm{tot,~max}}$', f_max_color='r'):
        self.plot_fw(ax_Fw)
        self.plot_exp_fw(ax_Fw)
        current_F = abs(self.final_plot_F_scale * self.beam_design.system_.F)
        w_SLS = self.beam_design.system_.L / 250
        if self.w_SLS:
            ax_Fw.plot([w_SLS, w_SLS], [0, 1.05 * current_F], linestyle='--', color='green')
            ax_Fw.annotate(r'$L/250$', xy=(1.02 * w_SLS, 0), color='green')
        ax_Fw.axhline(y=current_F, linestyle='--', color=f_max_color)
        ax_Fw.annotate(f_max_label + r'$ = $' + str(round(current_F, 2)) + ' ' + self.F_unit, xy=(0, 1.04 * current_F), color=f_max_color)

        if self.shear_force_can_be_calculated():
            self.get_nm_shear_force_capacity(should_print=True)

    def shear_force_can_be_calculated(self):
        # WHY does it have to be carbon here?
        # Because currently, shear calculation based on carbon empirical equations is implemented
        values = list(self.mc.cross_section_layout.items.values())
        return values[0].matmod == 'carbon' \
               and self.mc.cross_section_shape == 'rectangle'

    def plot_exp_fw(self, ax_Fw):
        for w, f in zip(self.w_exp_data, self.f_exp_data):
            ax_Fw.plot(w, f, label='exp deflection', lw=2)

    def plot_curvature_along_beam(self, ax_k):
        x = self.beam_design.system_.x
        kappa_x = self.get_kappa_x()
        # Plotting curvature
        ax_k.plot(x, kappa_x, color='black', label=r'$\kappa [\mathrm{mm}^{-1}]$')
        # Plotting rotation
        # ax_k.plot(x, self.get_phi_x()/1000, color='green', label='$\phi [rad]$')
        ax_k.fill_between(x, 0, kappa_x, color='gray', alpha=0.1)
        ax_k.invert_yaxis()
        ax_k.set_ylabel(r'$\kappa [\mathrm{mm}^{-1}]$')
        ax_k.set_xlabel(r'$x$')

    def plot_displacement_along_beam(self, ax_w):
        x = self.beam_design.system_.x
        w_x = self.get_w_x()
        ax_w.plot(x, w_x, color='blue', label='$w$ [mm]')
        ax_w.fill_between(x, 0, w_x, color='blue', alpha=0.1)
        ax_w.set_ylabel(r'$w [\mathrm{mm}]$')

    def plot_fw(self, ax_Fw, dashed=False, color='#1f77b4', label=None, lw=2):
        # TODO: expensive calculations for all displacements are running with each plot update to produce new
        #  load-displacement curve, this shouldn't be done for example when only the force has changed
        ax_Fw.set_xlabel(r'$w_\mathrm{max}$ [mm]')
        ax_Fw.set_ylabel(r'$F_{\mathrm{tot}}$ [' + self.F_unit + ']')
        F, w = self.get_Fw()
        ax_Fw.plot(w, self.final_plot_F_scale * F, '--' if dashed else '-', c=color, label='Sim.' if label is None else label, lw=lw)

    def get_nm_shear_force_capacity(self, should_print=False):
        # This works only for 4pb and dist and one layer of reinforcement
        system = self.beam_design.system
        mc = self.mc
        if system != '4pb' and system != 'simple_beam_dist_load':
            return
        if list(mc.cross_section_layout.items.values())[0].matmod != 'carbon':
            return

        z = list(mc.cross_section_layout.items.values())[0].z
        A_nm = list(mc.cross_section_layout.items.values())[0].A
        E_nm = list(mc.cross_section_layout.items.values())[0].matmod_.E
        d = mc.cross_section_shape_.H - z
        b = mc.cross_section_shape_.B
        L = self.beam_design.system_.L
        f_ck = mc.cs_design.matrix.compression_.f_cm - 8

        C_Rk_c = 0.219
        gamma = 1
        k = 1 / (1 + d / 200) ** 0.5
        rho_nm_l = A_nm / (b * d)
        lambda_ = L / (4 * d) if system == 'simple_beam_dist_load' else self.beam_design.system_.L_F / d
        exp = 1
        k_lambda = 1 + 2.825 * np.exp(-(lambda_) / 4.538) if system == 'simple_beam_dist_load' else 1 + 2.824 * np.exp(
            -(lambda_ - 1) / 4.538)
        V_Rm_c = C_Rk_c / gamma * k * k_lambda * (100 * E_nm / 200000 * rho_nm_l * f_ck) ** (1 / 3) * exp * b * d / 1000
        V_Rm_c_rounded = np.round(V_Rm_c, 2)
        if should_print:
            print('V_Rm_c = ' + str(V_Rm_c_rounded) + ' kN, Shear failure by F_tot_max = ' + str(2 * V_Rm_c_rounded) + ' kN')
        return 2 * V_Rm_c


class LoadDeflectionParamsStudy(ParametricStudy):

    def __init__(self, dp, show_sls_deflection_limit = False):
        self.dp = dp
        self.show_sls_deflection_limit = show_sls_deflection_limit

    def plot(self, ax, title, curve_label):
        ax.set_xlabel(r'$w_\mathrm{max}$ [mm]')
        ax.set_ylabel(r'$F$ [' + self.dp.F_unit + ']')
        F, w = self.dp.get_Fw()

        ax.plot(w, self.dp.final_plot_F_scale * F, label=curve_label, lw=2)

        if self.show_sls_deflection_limit:
            limit = self.dp.beam_design.system_.L/250
            ax.axvline(x=limit)
            ax.text(limit + 0.1, 0, 'L/250 = ' + str(self.dp.beam_design.system_.L/250), rotation=90)

        ax.set_title(title)
        ax.legend()
